@@ -2,5 +2,5 @@ from dotenv import load_dotenv
 load_dotenv()
 from .ai_agent import DMAIAgent
 from .async_ai_agent import DMAioAIAgent
-from .image_message_content_builder import ImageMessageContentBuilder
+from .openai_image_message_content import OpenAIImageMessageContent
 from .types import Message
